import asyncio
import json
import logging
import os

import requests
from fastapi import FastAPI, Request, HTTPException, Form
from fastapi.responses import Response
import httpx
import uvicorn
import urllib.parse

from kubernetes import client, config, watch
from pydantic import BaseModel

app = FastAPI()
logger = logging.getLogger('uvicorn.error')

# Configure target label and namespace; adjust as needed.
TARGET_PORT = os.getenv("TARGET_PORT", 80)
POD_LABEL_SELECTOR = os.getenv("POD_LABEL_SELECTOR", "name = wuenlp-server")
DEPLOYMENT_NAME = os.getenv("DEPLOYMENT_NAME", "wuenlp-server")
MIN_PODS = os.getenv("MIN_PODS", 1)
MAX_PODS = os.getenv("MAX_PODS", 10)
NAMESPACE = os.getenv("NAMESPACE", "web")
REFRESH_INTERVAL = os.getenv("REFRESH_INTERVAL", 10)  # seconds between pod list refresh

# Global shared state & lock.
# Mapping: pod IP -> {"busy": bool, "name": pod_name}
API_PODS = {}
global_lock = asyncio.Lock()


def load_k8s_config():
    """
    Attempt to load in-cluster configuration first,
    falling back to kubeconfig.
    """
    try:
        config.load_incluster_config()
    except Exception:
        config.load_kube_config()


def scale_deployment(new_replica_count: int):
    """
    Scale the deployment to the specified number of replicas.
    """
    apps_api = client.AppsV1Api()
    body = {'spec': {'replicas': new_replica_count}}
    apps_api.patch_namespaced_deployment_scale(
        name=DEPLOYMENT_NAME,
        namespace=NAMESPACE,
        body=body
    )
    print(f"Deployment scaled to {new_replica_count} replicas.")


def delete_pod(pod_name: str):
    """
    Delete a specific pod by its Kubernetes name.
    """
    v1 = client.CoreV1Api()
    try:
        # V1DeleteOptions can be customized (e.g., grace period, propagation policy)
        v1.delete_namespaced_pod(
            name=pod_name,
            namespace=NAMESPACE,
            body=client.V1DeleteOptions()
        )
        print(f"Pod {pod_name} deletion initiated.")
    except Exception as e:
        print(f"Error deleting pod {pod_name}: {e}")


async def update_api_pods():
    """
    Background task that periodically queries the Kubernetes API
    to update the list of API pods based on a label selector.
    """
    load_k8s_config()  # Ensure configuration is loaded
    v1 = client.CoreV1Api()
    while True:
        try:
            pod_list = v1.list_namespaced_pod(
                namespace=NAMESPACE,
                label_selector=POD_LABEL_SELECTOR
            )
            new_pods = {}
            for pod in pod_list.items:
                if pod.status.phase != "Running":
                    continue
                pod_ip = pod.status.pod_ip
                if not pod_ip:
                    continue
                # Retain the busy state if this pod was already tracked.
                busy_state = API_PODS.get(pod_ip, {}).get('busy', False)
                new_pods[pod_ip] = {"busy": busy_state, "name": pod.metadata.name}
            async with global_lock:
                API_PODS.clear()
                API_PODS.update(new_pods)
        except Exception as e:
            print(f"Error updating pods: {e}")
        await asyncio.sleep(REFRESH_INTERVAL)


async def get_free_pod() -> (str, str):
    """
    Wait for a free pod to be available. If no free pod exists and the
    current count is below MAX_PODS, scale up the deployment to spawn an additional pod.

    Returns:
        A tuple containing the pod IP and the pod name.
    """
    while True:
        async with global_lock:
            for pod_ip, info in API_PODS.items():
                if not info["busy"]:
                    API_PODS[pod_ip]["busy"] = True
                    return pod_ip, info["name"]
            # If no free pod is available and we haven't hit MAX_PODS, scale up.
            if len(API_PODS) < MAX_PODS:
                new_replica_count = len(API_PODS) + 1
                print(f"All pods busy. Scaling up deployment to {new_replica_count} replicas.")
                try:
                    scale_deployment(new_replica_count)
                except Exception as e:
                    print(f"Error scaling up deployment: {e}")
        await asyncio.sleep(10)


async def mark_pod_free(pod_ip: str, pod_name: str):
    """
    Mark the specified pod as free. Also, if extra pods exist (above MIN_PODS),
    trigger a scale down by explicitly deleting a free pod.

    Args:
        pod_ip: The IP of the pod being freed.
        pod_name: The Kubernetes name of the pod being freed.
    """
    async with global_lock:
        if pod_ip in API_PODS:
            API_PODS[pod_ip]["busy"] = False

        # Gather all free pods.
        free_pods = [(ip, info["name"]) for ip, info in API_PODS.items() if not info["busy"]]
        current_count = len(API_PODS)
        if free_pods and current_count > MIN_PODS:
            new_replica_count = current_count - 1
            # Pick a free pod for deletion.
            target_ip, target_name = free_pods[0]
            print(f"Scaling down deployment to {new_replica_count} replicas by deleting pod {target_name}.")
            try:
                scale_deployment(new_replica_count)
                delete_pod(target_name)
                # Remove the deleted pod from tracking.
                del API_PODS[target_ip]
            except Exception as e:
                print(f"Error scaling down deployment: {e}")


@app.on_event("startup")
async def startup_event():
    # Start the background task to update API pods.
    asyncio.create_task(update_api_pods())


@app.api_route("/{path:path}", methods=["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "HEAD"])
async def proxy(path: str, request: Request):
    logger = app.logger if hasattr(app, "logger") else print
    logger(f"Received request for path: /{path}")

    # Wait until we assign a free backend pod.
    pod_ip, pod_name = await get_free_pod()
    logger(f"Selected pod: {pod_ip}, {pod_name}")

    # Construct the target URL.
    target_url = f"http://{pod_ip}:{TARGET_PORT}/{path}"
    # Also preserve any query parameters.
    query_string = request.url.query
    await asyncio.sleep(20)
    try:
        # Read the request body as bytes (this will work for any content type).
        body_bytes = await request.body()

        # Capture the original headers. You might want to filter out any headers that
        # are restricted or need to be rewritten (like the Host header).
        orig_headers = dict(request.headers)
        # Optionally remove/change headers if needed:
        orig_headers.pop("host", None)

        # Use httpx to perform the same request to the target pod.
        async with httpx.AsyncClient(timeout=None) as client:
            response = await client.request(
                method=request.method,
                url=target_url,
                params=urllib.parse.parse_qs(query_string),
                content=body_bytes,
                headers=orig_headers
            )

        # Prepare response headers.
        response_headers = {
            key: value
            for key, value in response.headers.items()
            if key.lower() in ["content-type", "content-disposition"]
        }
        # Return the response from the backend API exactly as received.
        return Response(
            content=response.content,
            status_code=response.status_code,
            headers=response_headers
        )
    except Exception as err:
        raise HTTPException(status_code=500, detail=str(err))
    finally:
        # Release the pod resource
        await mark_pod_free(pod_ip, pod_name)


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)
